#
#Licensed to the Apache Software Foundation (ASF) under one
#or more contributor license agreements.  See the NOTICE file
#distributed with this work for additional information
#regarding copyright ownership.  The ASF licenses this file
#to you under the Apache License, Version 2.0 (the
#"License"); you may not use this file except in compliance
#with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing,
#software distributed under the License is distributed on an
#"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#KIND, either express or implied.  See the License for the
#specific language governing permissions and limitations
#under the License.
#
require "logstash/outputs/base"
require "logstash/namespace"
require "logstash/environment"
require "fileutils"
require "thread"

jar_path=File.expand_path(File.join(File.dirname(__FILE__), "../../.."))
LogStash::Environment.load_runtime_jars! File.join(jar_path, "vendor")

# Datahub output plugin
class LogStash::Outputs::Datahub < LogStash::Outputs::Base
  declare_threadsafe!

  config_name "datahub"

  # datahub access id
  config :access_id, :validate => :string, :required => true

  # datahub access key
  config :access_key, :validate => :string, :required => true

  # datahub service endpoint
  config :endpoint, :validate => :string, :required => true

  # datahub project name
  config :project_name, :validate => :string, :required => true

  # datahub topic name
  config :topic_name, :validate => :string, :required => true

  # 重试次数，-1为无限重试、0为不重试、>0表示需要有限次数
  config :retry_times, :validate => :number, :required => false, :default => -1

  # 重试周期，下一次重试的间隔，单位为秒
  config :retry_interval, :validate => :number, :required => false, :default => 5

  # 按照指定字段的值计算hash，依据于该hash值落某个shard
  config :shard_keys, :validate => :array, :required => false, :default => []

  # 指定数据落指定的shard
  config :shard_id, :validate => :string, :required => false, :default => ""

  #   # 提交的列名，用户可以配置topic的列，采集部分列或者全部列
  #   # 默认为空数组，表示按照topic的顺序及全字段提交
  #   # 另外：列的配置不用保序，但是要求该字段在topic的schema中存在
  #   config :column_names, :validate => :array, :required => false, :default => []

  # 当出现脏数据时，是否继续写入
  # 当开启该开关，必须指定@dirty_data_file文件
  config :dirty_data_continue, :validate => :boolean, :required => false, :default => false

  # 脏数据文件名称，当数据文件名称，在@dirty_data_continue开启的情况下，需要指定该值
  # 特别注意：脏数据文件将被分割成两个部分.part1和.part2，part1作为更早的脏数据，part2作为更新的数据
  config :dirty_data_file, :validate => :string, :required => false

  # 脏数据文件的最大大小，该值保证脏数据文件最大大小不超过这个值，目前该值仅是一个参考值
  config :dirty_data_file_max_size, :validate => :number, :required => false, :default => 50024000

  # 数据传输压缩方式选择，目前支持deflate, lz4格式 
  config :compress_method, :validate => :string, :required => false, :default => ""

  # 该值内部使用，不提供配置
  # 分发shard的游标
  attr_accessor :shard_cursor

  # Shard cursor lock
  @@shard_lock = Mutex.new

  # 写文件锁
  @@file_lock = Mutex.new

  DatahubPackage = com.aliyun.datahub

  public
  def register
    begin
      @account = DatahubPackage.auth.AliyunAccount::new(@access_id, @access_key)
      @conf = DatahubPackage.DatahubConfiguration::new(@account, @endpoint)
      if @compress_method == "deflate" || @compress_method == "lz4"
        @compression_format = DatahubPackage.model.compress.CompressionFormat.fromValue(@compress_method)
        @conf.setCompressionFormat(@compression_format)
      end

      @client = DatahubPackage.DatahubClient::new(@conf)
      @project = DatahubPackage.wrapper.Project::Builder.build(@project_name, @client)
      @topic = @project.getTopic(@topic_name)
      @shard_cursor = 0

      @shards = get_active_shards(@topic.listShard())
      @shard_count = @shards.size()

      result = @client.getTopic(@project_name, @topic_name)
      @schema = result.getRecordSchema()
      fields = @schema.getFields()
      @columns_size = fields.size
      @columnnames = []
      for i in 0...@columns_size
        @columnnames.push(fields[i].getName())
      end
      @columntypes = []
      for i in 0...@columns_size
        @columntypes.push(fields[i].getType())
      end

      # 前置校验参数
      check_params()

      if @shard_count == 0
        @logger.error "No active shard available, please check"
        raise "No active shard available, please check"
      end

      @logger.info "Init datahub success!"
    rescue => e
      @logger.error "Init failed!"  + e.message + " " + e.backtrace.inspect.to_s
      raise e
    end
  end # def register

  def check_params()
    # 如果shard_id配置了，则检查该shard是否ok
    if !@shard_id.empty?
      valid = false
      for i in 0...@shards.size
        shard_entry = @shards[i]
        if shard_entry.getShardId() == @shard_id && shard_entry.getState() == DatahubPackage.model.ShardState::ACTIVE
          valid = true
        end
      end
      if (!valid)
        @logger.error "Config shard_id not exists or state not active, check your config"
        raise "Config shard_id not exists or state not active, check your config"
      end
    end

    # 检查shard_keys字段是否合法
    if @shard_keys.size > 0
      for i in 0...@shard_keys.size
        shard_key = @shard_keys[i]
        if !@schema.containsField(shard_key)
          @logger.error "Config shard_keys contains one or one more unknown field, check your config"
          raise "Config shard_keys contains one or one more unknown field, check your config"
        end
      end
    end

    # 配置了脏数据继续，必须指定脏数据文件
    if @dirty_data_continue
      if @dirty_data_file.to_s.chomp.length == 0
        raise "Dirty data file path can not be empty"
      end
    end

  end

  # 检查并设置数据到entry中
  # 如果解析数据异常，则数据落脏数据文件
  def check_and_set_data(entry, field_type, index, event_map, column_name)
    data = event_map[column_name]
    begin
      if field_type == DatahubPackage.common.data.FieldType::STRING
        entry.setString(index, data.to_s)
      elsif field_type == DatahubPackage.common.data.FieldType::BIGINT
        entry.setBigint(index, java.lang.Long.parseLong(data.to_s))
      elsif field_type == DatahubPackage.common.data.FieldType::DOUBLE
        entry.setDouble(index, java.lang.Double.parseDouble(data.to_s))
      elsif field_type == DatahubPackage.common.data.FieldType::BOOLEAN
        entry.setBoolean(index, java.lang.Boolean.parseBoolean(data.to_s))
      elsif field_type == DatahubPackage.common.data.FieldType::TIMESTAMP
        entry.setTimeStamp(index, java.lang.Long.parseLong(data.to_s))
      else
        raise "Unknown schema type of data"
      end
      return true
    rescue => e
      @logger.error "Parse data: " + column_name + "[" + data + "] failed, " + e.message
      # 数据格式有异常，根据配置参数确定是否续跑
      if !@dirty_data_continue
        @logger.error "Dirty data found, exit process now."
        puts "Dirty data found, exit process now."
        Process.exit(1)
        # 忽略的异常数据直接落文件
      else
        write_as_dirty_data(event_map)
      end
      return false
    end
  end

  # 脏数据文件处理
  def write_as_dirty_data(event_amp)
    dirty_file_part1_name = @dirty_data_file + ".part1"
    dirty_file_part2_name = @dirty_data_file + ".part2"

    # 加锁写入
    @@file_lock.synchronize {
      dirty_file_part2 = File.open(dirty_file_part2_name, "a+")
      dirty_file_part2.puts(event_amp.to_s)
      dirty_file_part2.close
      if File.size(dirty_file_part2_name) > @dirty_data_file_max_size / 2
        # .part1, .part2分别存储数据
        # 旧数据落part1，新的数据落part2
        FileUtils.mv(dirty_file_part2_name, dirty_file_part1_name)
      end
    }
  end

  def get_active_shards(shards)
    active_shards = []
    for i in 0...shards.size
      entry = shards.get(i)
      if entry.getState() == DatahubPackage.model.ShardState::ACTIVE
        active_shards.push(entry)
      end
    end
    return active_shards
  end

  def get_next_shard_id()
    if !@shard_id.empty?
      return @shard_id
      # 否则轮询写入shard
    else
      idx = 0
      @@shard_lock.synchronize {
        idx = @shard_cursor % @shard_count
        @shard_cursor = idx + 1
      }
      shard_id = @shards[idx].getShardId()
      return shard_id
    end
  end

  def multi_receive(event_list)
    retry_count = 0
    begin
      entries = []
      shard_id = get_next_shard_id()

      event_list.each do |event|
        if event == LogStash::SHUTDOWN
          return
        end
        event_map = event.to_hash

        entry = DatahubPackage.model.RecordEntry::new(@schema)
        entry.putAttribute("srcId", event_map["host"].to_s)
        entry.putAttribute("ts", event_map["@timestamp"].to_s)
        entry.putAttribute("version", event_map["@version"].to_s)
        entry.putAttribute("srcType", "log")

        is_data_valid = false
        for i in 0...@columns_size do
          column_name = @columnnames[i]
          column_type = @columntypes[i]
          value = event_map[column_name]
          if value != nil
            is_data_valid = check_and_set_data(entry, column_type, i, event_map, column_name)
            break if !is_data_valid
          end
        end

        if is_data_valid
          if @shard_keys.size > 0
            hash_string = ""
            for i in 0...@shard_keys.size
              shard_key = @shard_keys[i]
              if event_map[shard_key] != nil
                hash_string += event_map[shard_key].to_s + ","
              end
            end
            hashed_value = java.lang.String.new(hash_string).hashCode()
            entry.setPartitionKey(hashed_value)
          else
            entry.setShardId(shard_id)
          end
          entries.push(entry)
        end
      end

      # puts "total: " + entries.size.to_s

      # 提交列表必须有数据
      if entries.size > 0
        put_result = @client.putRecords(@project_name, @topic_name, entries)
        if put_result.getFailedRecordCount() > 0
          @logger.info "Put " + put_result.getFailedRecordCount().to_s + " records to datahub failed, total " + entries.size().to_s
          sleep @retry_interval
          entries = put_result.getFailedRecords()
          raise "Write to datahub failed: " + entries.size.to_s
        else
          @logger.info "Put data to datahub success, total " + entries.size().to_s
        end
      end

    rescue DatahubPackage.exception.DatahubServiceException => e
      @logger.error "Flush data exception: " + e.message #+ " " + e.backtrace.inspect.to_s
      # shard的状态改变，需要重新加载shard
      if e.getErrorCode() == "InvalidShardOperation"
        @shards = get_active_shards(@topic.listShard())
        @shard_count = @shards.size()

        if @shard_count == 0
          @logger.error "No active shard available, please check"
        end
      elsif e.getErrorCode() == nil
        sleep @retry_interval
      end
      retry_count += 1
      @logger.warn "Now retry: " + retry_count.to_s
      retry
    rescue => e
      @logger.error "Flush data exception: " + e.message + " " + e.backtrace.inspect.to_s

      # 无限重试
      if @retry_times < 0
        retry_count += 1
        @logger.warn "Now retry: " + retry_count.to_s
        # puts "Now retry..."
        sleep @retry_interval
        retry
      elsif @retry_times == 0
        @logger.error "Retry not work, now exit"
        Process.exit(1)
        # 继续重试
      elsif @retry_times > 0
        retry_count += 1
        if retry_count > @retry_times
          @logger.warn "Retry over: " + @retry_times.to_s
          Process.exit(1)
        end
        @logger.warn "Now retry..."
        sleep @retry_interval
        retry
      end
    end
  end # def multi_receive

end # class LogStash::Outputs::Datahub
